
(* C89 AST extended with assume statements *)

type symbol = Types.symbol
type const = Expr.const

type program = external_declaration list
and external_declaration = FunctionDefinition of function_definition | Declaration of declaration
and function_definition = symbol * declaration_specifiers * declarator * statement
and declaration = declaration_specifiers * (symbol * declarator * init option) list

and declaration_specifiers = storage_class_specifier list * type_specifier list * type_qualifier list
and storage_class_specifier = Typedef | Extern | Static | Auto | Register
and type_qualifier = Volatile | Const
and type_specifier =
  | Void
  | Char
  | Short
  | Int
  | Long
  | Float
  | Double
  | Signed
  | Unsigned
  | Struct of symbol option * (struct_declaration list) option
  | Union of symbol option * (struct_declaration list) option
  | Enum of symbol option * (enumerator list) option
  | UserType of symbol

and declarator =
  | Pointer of type_qualifier list * declarator
  | Array of declarator * expression option
  | Function of declarator * function_parameter list
  | Scalar
and function_parameter = Parameter of parameter | Ellipsis
and parameter = symbol option * declaration_specifiers * declarator
and init = ScalarValue of expression | ArrayValue of init list

and struct_declaration = type_specifier list * type_qualifier list * struct_declarator list
and struct_declarator = symbol option * declarator * expression option
and enumerator = symbol * expression option

and compound_statement = declaration list * statement list
and label = Label of symbol | Case of expression | Default
and statement =
  | Nop
  | Assume of expression
  | LabelledStatement of label * statement
  | Compound of compound_statement
  | Expression of expression
  | If of expression * statement * statement
  | Switch of expression * statement
  | While of expression * statement
  | Do of expression * statement
  | For of expression option * expression option * expression option * statement
  | Continue
  | Goto of symbol
  | Break
  | Return of expression option

and type_name = type_specifier list * type_qualifier list * declarator
and expression = 
  | Variable of symbol
  | Constant of const
  | String of string
  | ArrayAccess of expression * expression
  | FunctionCall of expression * expression list
  | Selection of expression * symbol
  | PointerSelection of expression * symbol
  | PreIncrementation of expression
  | PreDecrementation of expression
  | PostIncrementation of expression
  | PostDecrementation of expression
  | Adress of expression
  | Dereferenciation of expression
  | Neg of expression
  | BinaryNot of expression
  | LogicalNot of expression
  | SizeofExpression of expression
  | SizeofType of type_name
  | Cast of type_name * expression
  | Mul of expression * expression
  | Div of expression * expression
  | Mod of expression * expression
  | Add of expression * expression
  | Sub of expression * expression
  | ShiftLeft of expression * expression
  | ShiftRight of expression * expression
  | Lower of expression * expression
  | Greater of expression * expression
  | LowerEq of expression * expression
  | GreaterEq of expression * expression
  | Eq of expression * expression
  | Neq of expression * expression
  | BinaryAnd of expression * expression
  | BinaryOr of expression * expression
  | BinaryXor of expression * expression
  | LogicalAnd of expression * expression
  | LogicalOr of expression * expression
  | Conditional of expression * expression * expression
  | Assign of expression * expression
  | MulAssign of expression * expression
  | DivAssign of expression * expression
  | ModAssign of expression * expression
  | AddAssign of expression * expression
  | SubAssign of expression * expression
  | LeftShiftAssign of expression * expression
  | RightShiftAssign of expression * expression
  | AndAssign of expression * expression
  | XorAssign of expression * expression
  | OrAssign of expression * expression
  | Sequence of expression * expression
