/* % $Id: parser.mly,v 1.1.1.1 2003/04/18 19:20:36 jeannet Exp $ i*/
/* modified by Lg on april 2006 for the use of pfcalc-accel */
/* modified by Valentin Perrelle on april 2009 for use in enkidu */
/* Parser of the fast langage */

%{
open FastModel
open Types
open Expr
open Expr.StdExpression
%}

/* %============================================================== */
/* \section{Tokens} */
/* %============================================================== */

%token TK_EOF

/* dlimiteurs */
%token TK_LPAREN TK_RPAREN TK_LBRACE TK_RBRACE TK_LBRACKET TK_RBRACKET
%token TK_COMMA TK_SEMI
%token TK_PRIME
%token TK_COLON
%token TK_VBAR TK_RARROW

/* boolean operations */
%token TK_TRUE TK_FALSE
%token TK_NOT TK_OR TK_AND TK_NEQUAL TK_EQUAL TK_SHARP TK_IMPLY TK_IN

/* arithmetic operations */
%token TK_MUL TK_DIV TK_PLUS TK_MINUS

/* special values */
%token TK_UNDETERMINISTIC

/* oprations mixtes */
%token TK_GREATEREQUAL TK_LESSEQUAL TK_GREATER TK_LESS TK_ASSIGN TK_DISEQUAL

/* control instructions */
%token TK_IF TK_THEN TK_ELSE

/* declarations */
%token TK_TYPEDEF TK_ENUM TK_BOOL TK_INT TK_REAL TK_UINT TK_SINT
%token TK_INVARIANT TK_BLOCKED TK_ANYTHING TK_INITIAL TK_SPACE TK_FINAL TK_ASSERTION
%token TK_INPUT
%token TK_LOCAL
%token TK_STATE
%token TK_CONST
%token TK_TRANS TK_DEFINITION
%token TK_CONTINUOUS
%token TK_CONTROL TK_AUTOMATON TK_LOCATION TK_EDGE TK_PRIORITY
%token <int> TK_NUM
%token <Rational.t> TK_RAT
%token <Types.symbol> TK_ID


/* associativity */
%nonassoc TK_EQUIV
%right TK_IMPLY
%left TK_OR TK_OROR
%left TK_AND TK_ANDAND
%nonassoc TK_NOT TK_EXCLA
%nonassoc TK_GREATER TK_GREATEREQUAL TK_LESS TK_LESSEQUAL TK_EQUAL TK_NEQUAL TK_DISEQUAL
%left TK_PLUS TK_MINUS
%left TK_MUL TK_DIV

/* fast */
%token TK_MODEL
%token TK_STRATEGY
%token TK_STRAT
%token TK_VAR
%token TK_ARRAY
%token TK_STATES
%token TK_FROM
%token TK_TO
%token TK_GUARD
%token TK_ACTION
%token TK_REGION
%token TK_ANDAND
%token TK_OROR
%token TK_EXCLA
%token TK_PRE
%token TK_POST
%token TK_SETMAXSTATE
%token TK_SETMAXACC
%token TK_PRINT
%token TK_BOOLEAN
%token TK_STRATSUBSET 
%token TK_STRATEQSET
%token TK_STRATISEMPTY
%token TK_STRATTRANS
%token TK_EXISTS
%token TK_STRATTRUE
%token TK_STRATFALSE
%token TK_ENDIF
%token TK_EQUIV
%token TK_DOUBLEQUOTE

%start prog

%type <FastModel.program> prog
%%


prog: model strat
{ 
  { 
    model = $1;
    strategy = $2
  }
}

/*MODEL*/
model : TK_MODEL id TK_LBRACE vars arrays states transitions TK_RBRACE
{
  { 
    idmodel = $2;
    vars = $4;
    arrays = $5;
    states = $6;
    transitions = $7;
  }
}


vars : TK_VAR idlist TK_SEMI
{
  $2
}

arrays : 
| {[]} /* no arrays */
| TK_ARRAY idlist TK_SEMI { $2 }

id_: 
| TK_ID { Variable($1) }
| TK_ID TK_LBRACKET numexpr_ TK_RBRACKET { Array($1,$3) } 

id: TK_ID
{
 $1 
}

idlist:
id { [$1] }
| idlist TK_COMMA id { $3::$1 }

states : TK_STATES idlist TK_SEMI
{
  $2
}


transitions : translist 
{
  $1
}

translist :
trans { [$1] }
| translist trans {$2::$1}


trans : TK_TRANS id TK_ASSIGN TK_LBRACE intrans TK_RBRACE TK_SEMI
{
  let (fromid,toid,guard,action) = $5 in
    {
      ident = $2;
      fromid = fromid;
      toid = toid;
      guard = guard;
      actions = action
    }
}


intrans : TK_FROM TK_ASSIGN id TK_SEMI TK_TO TK_ASSIGN id TK_SEMI TK_GUARD TK_ASSIGN guard TK_SEMI intransaction
{
  ($3,$7,$11,$13)
}

intransaction : 
| {[]} /*action ne rien faire */
| TK_ACTION TK_ASSIGN action TK_SEMI {$3}

action : 
| affectation {[$1]}
| action TK_COMMA affectation {$1@[$3]}

affectation : id_ TK_PRIME TK_EQUAL expr
{
  (NumericalAssign ($1,$4))
}


/*Strategy*/

strat : TK_STRATEGY id TK_LBRACE instructionlist TK_RBRACE
{
  {
    idstrat = $2;
    instlist = $4;
  }
}


instructionlist : 
instruction {[$1]}
| instructionlist instruction {$2::$1} 


instruction : 
TK_REGION id TK_ASSIGN region TK_SEMI{Declaration(Region($2,$4))}
| TK_STRATTRANS id TK_ASSIGN  idlistbraq TK_SEMI {Declaration(Transitions($2,$4))}
/*| TK_BOOLEAN id TK_ASSIGN boolfastexpr TK_SEMI {Declaration(Types.Boolean($2,$4))} */
| TK_SETMAXSTATE  TK_LPAREN integ TK_RPAREN TK_SEMI{Proc(1,$3)} 
| TK_SETMAXACC TK_LPAREN integ TK_RPAREN TK_SEMI {Proc(2,$3)} 
| TK_PRINT TK_LPAREN chaine  TK_RPAREN TK_SEMI{Print} /*ignore*/
| TK_IF boolfastexpr TK_THEN instructionlist suite {Ite($2,$4,$5)} /*ignore*/

suite :
TK_ENDIF {[]}
| TK_ELSE instructionlist TK_ENDIF {$2}

chaine :
id {}
| TK_DOUBLEQUOTE idandspace TK_DOUBLEQUOTE {}

idandspace : 
id {}
| id idandspace {}


idlistbraq :
id {[$1]}
| TK_LBRACE idlist TK_RBRACE{ $2 }



integ:
| TK_NUM {$1}

region : 
regionterm {$1}
| region TK_ANDAND region {Inter($1,$3)}
| region TK_OROR region {Union($1,$3)}
| TK_EXCLA region {Complement($2)}
| TK_LPAREN region TK_RPAREN {$2}
| TK_PRE TK_MUL TK_LPAREN region TK_COMMA  idlist TK_RPAREN {(PreStar($4,$6,0))}
| TK_PRE TK_MUL TK_LPAREN region TK_COMMA  idlist TK_COMMA  integ TK_RPAREN {PreStar($4,$6,$8)}
| TK_POST TK_MUL TK_LPAREN region TK_COMMA  idlist TK_RPAREN {(PostStar($4,$6,0))}
| TK_POST TK_MUL TK_LPAREN region TK_COMMA  idlist TK_COMMA integ TK_RPAREN {PostStar($4,$6,$8)}
		 
regionterm : 
id {RegiontermI($1)}
| TK_LBRACE guard TK_RBRACE {RegiontermF($2)}

boolfastexpr :
| id {Cstfast(Id($1))}
| TK_TRUE {Cstfast(Cst_true)}
| TK_FALSE {Cstfast(Cst_false)}
| TK_STRATSUBSET TK_LPAREN region TK_COMMA region TK_RPAREN {SpecialSubset($3,$5)}
| TK_STRATEQSET TK_LPAREN region TK_COMMA region TK_RPAREN {SpecialEqset($3,$5)}
| TK_STRATISEMPTY  TK_LPAREN region TK_RPAREN{SpecialIsEmpty($3)}
| boolfastexpr TK_AND boolfastexpr {BoolAnd($1,$3)}
| boolfastexpr TK_OR boolfastexpr {BoolOr($1,$3)}
| boolfastexpr TK_EQUAL boolfastexpr {BoolEq($1,$3)}
| TK_EXCLA TK_LPAREN boolfastexpr TK_RPAREN{Negation($3)}
 | TK_LPAREN boolfastexpr TK_RPAREN {$2} 


/*guards are formula over numerical constraints*/
/*non primed formula*/

guard : 
 | TK_LPAREN guard TK_RPAREN {$2}
 | TK_TRUE {True}
 | TK_FALSE {False}
 | guard TK_ANDAND guard {And($1,$3)}
 | guard TK_OROR guard {Or($1,$3)}
 | guard TK_IMPLY guard {Imply($1,$3)}
 | guard TK_EQUIV guard {And(Imply($1,$3),Imply($3,$1))}
 | TK_EXCLA  TK_LPAREN guard TK_RPAREN {Not($3)}
 | numexpr TK_GREATEREQUAL numexpr {Comparison(GreaterEq($1,$3))}
 | numexpr TK_GREATER numexpr {Comparison(Greater($1,$3))}
 | numexpr TK_LESSEQUAL numexpr {Comparison(GreaterEq($3,$1))}
 | numexpr TK_LESS numexpr {Comparison(Greater($3,$1))}
 | numexpr TK_EQUAL numexpr {Comparison(Eq($1,$3))}
 | numexpr TK_DISEQUAL numexpr {Comparison(Neq($1,$3))}

numexpr : 
  expr6 { $1 }

numexpr_ :
  expr6_ { $1 }

expr6_:
  expr8_ { $1 }
| expr6_ TK_PLUS expr8_ {Plus($1,$3)}
| expr6_ TK_MINUS expr8_ {Minus($1,$3)}

expr8_:
| numcst   {Value(Const($1))}
| id { Value(Variable($1)) }
| numcst id { Mul(Value(Const($1)),Value(Variable($2))) }
| TK_LPAREN expr6_ TK_RPAREN { $2 }
| TK_MINUS expr8_ { Minus(Value(Const(Integer(0))),$2) }


/* %======================================================================== */
/* \subsection{Standard expressions} */
/* %======================================================================== */

/* Expressions are Boolean, linear affine expression, or special
expression to define the behaviour of continuous evolving
variables. Type checking is not done in that place, so the syntax is
quite free.  */

expr:
  expr6 { $1 }
/*
  expr0 { $1 }
| TK_IF expr TK_THEN expr TK_ELSE expr { If($2,$4,$6) }
expr0:
  expr1 { $1 }
| expr1 TK_IMPLY expr0 { Imply($1,$3) }
expr1:
  expr2 { $1 }
| expr1 TK_OR expr2   { Binop(Or,$1,$3) }
expr2:
  expr3 { $1 }
| expr2 TK_AND expr3  { Binop(And,$1,$3) }
expr3:
  expr4 { $1 }
| expr4 TK_NEQUAL expr4  { Monop(Not,Binop(Eq,$1,$3)) }
| expr4 TK_EQUAL expr4   { Binop(Eq,$1,$3) }
| expr8 TK_IN TK_LBRACE list_expr TK_RBRACE  { In($1,$4) }
expr4:
  expr5 { $1 }
| TK_NOT expr4       { Monop(Not,$2) }
| TK_SHARP TK_LPAREN list_expr TK_RPAREN { Excl($3) }
expr5:
  expr6 { $1 }
| expr6 TK_GREATEREQUAL expr6 { Monop(SupEg,Binop(Moins,$1,$3)) }
| expr6 TK_GREATER expr6 { Monop(Sup,Binop(Moins,$1,$3)) }
| expr6 TK_LESSEQUAL expr6 { Monop(SupEg,Binop(Moins,$3,$1)) }
| expr6 TK_LESS expr6 { Monop(Sup,Binop(Moins,$3,$1)) }
*/
expr6:
  expr7 { $1 }
| expr6 TK_PLUS expr7 { Plus($1,$3) }
| expr6 TK_MINUS expr7 { Minus($1,$3) }
expr7:
  expr8 { $1 }
| expr7 TK_MUL expr8 { Mul($1,$3) }
| expr7 TK_DIV expr8 { Div($1,$3) }
expr8:
| numcst   { Value (Const($1)) }
/*
| bint TK_LPAREN TK_NUM TK_RPAREN { Cst(BInt($1,$3)) }
| bint TK_LPAREN TK_MINUS TK_NUM TK_RPAREN { Cst(BInt($1,-$4)) }
*/
| id_ { Value($1) }
| numcst id { Mul(Value(Const($1)),Value(Variable($2))) }
| TK_LPAREN expr6 TK_RPAREN { $2 }
| TK_MINUS expr8 { Neg($2) }
| TK_UNDETERMINISTIC { Value(Unknown) }


numcst:
  TK_NUM { Integer($1) }
| TK_RAT { Rational($1) }
/*
list_expr:
  expr { [$1] }
| list_expr TK_COMMA expr { $3::$1 }
*/
/*
bint:
| TK_UINT TK_LBRACKET TK_NUM TK_RBRACKET
{ 
  if $3 < 0 then
    raise (ParsingError (Print.sprintf "Error: in type uint[x], x should be positive"))
  ;
  (false,$3)
}
| TK_SINT TK_LBRACKET TK_NUM TK_RBRACKET
{ 
  if $3 < 0 then
    raise (ParsingError (Print.sprintf "Error: in type sint[x], x should be positive"))
  ;
  (true,$3)
}
*/